
Require Import Terms.

Definition tForce (tsum : Tm) : Tm :=
  let x := 0 in
  tMatch tsum
    x (TVar x)
      (TThrow x).

Definition tForceTag (b : BOp) (t : Tm) : Tm :=
  let x := fresh (fv_tm t) in
  TLet x (tForce t)
    (match b with
     | BEq =>
       tIf (tEq (TTag x) (tCTag TagFun))
         (tThrow (tCExcp eType))
         (tIf (tEq (TTag x) (tCTag TagSum))
           (tThrow (tCExcp eType))
           (TVar x))
     | _ =>
       tIf (tEq (TTag x) (tCTag TagLab))
         (TVar x)
         (tThrow (tCExcp eType))
    end).

Fixpoint encode (t : Tm) : Tm :=
  match t with
  | TVar x => TVar x
  | TConst k => tInl (TConst k)
  | TLet x t1 t2 => TLet x (encode t1) (encode t2)
  | TAbs x t => tInl (TAbs x (encode t))
  | TApp x y =>
      (* not sure if it's possible to do this without type tags
         - there are 2 more uses of tHasTag below though *)
      let x' := fresh [x, y] in
      tSBind (TVar x) x'
        (tHasTag TagFun (TVar x')
          (TApp x' y))
  | TInx d x => tInl (TInx d x)
  | TMatch x x1 t1 t2 =>
      let x' := fresh (x :: x1 :: fv_tm t1 ++ fv_tm t2) in
      tSBind (TVar x) x'
        (tHasTag TagSum (TVar x')
          (TMatch x' x1 (encode t1) (encode t2)))
  | TTag x =>
      tBracket (TLabelOf x) (tTag (tForce (TVar x)))
  | TBOp BJoin x y =>
      tBracket (tJoin (TLabelOf x) (TLabelOf y))
        (tBOp BJoin (tForceTag BJoin (TVar x))
                    (tForceTag BJoin (TVar y)))
  (* additional massaging for BEq and BCmp
     since they return booleans encoded as sums *)
  | TBOp b x y =>
      let z := 0 in
      tBracket (tJoin (TLabelOf x) (TLabelOf y))
        (* playing the same trick as for toSum *)
        (tMatch (tBOp b (tForceTag b (TVar x))
                        (tForceTag b (TVar y)))
          z (tInl (TInx DLeft z))
            (tInr (TInx DLeft z)))
  | TBracket x t =>
      let x' := fresh (x :: fv_tm t) in
      tSBind (TVar x) x'
        (tHasTag TagLab (TVar x')
          (TBracket x' (tForce (encode t))))
  | TLabelOf x => tInl (TLabelOf x)
  | TGetPc => tInl TGetPc
  | TMkNav x =>
      tBracket (TLabelOf x)
        (tThrow (tForce (TVar x)))
  | TToSum x =>
      let y := 0 in
      (tBracket (TLabelOf x)
        (TMatch x
          y (tInl (TInx DLeft y))
            (tInr (TInx DLeft y))))
  | TThrow _ => tCUnit (* not part of LambdaNav *)
  | TCatch _ _ _ => tCUnit (* not part of LambdaNav *)
  end.

Require LNaVSyntax.
Require LNaVBigStep.
Require LBracketSyntax.
Require LThrowBigStep.

(* Require without Import doesn't bring in tactics,
   and they also don't seem accessible with LBracketSyntax. notation.
   So here goes some duplication *)
Tactic Notation "evalLN_cases" tactic(first) ident(c) :=
  first; [
  (* perl -ne 'print "    Case_aux c \"$1\" |\n" if /\| (eval_\w+)\s/' < LNaVBigStep.v *)
    Case_aux c "eval_var" |
    Case_aux c "eval_const" |
    Case_aux c "eval_let" |
    Case_aux c "eval_abs" |
    Case_aux c "eval_app" |
    Case_aux c "eval_app_no_abs" |
    Case_aux c "eval_inx" |
    Case_aux c "eval_match" |
    Case_aux c "eval_match_no_sum" |
    Case_aux c "eval_tag" |
    Case_aux c "eval_bop" |
    Case_aux c "eval_bracket" |
    Case_aux c "eval_bracket_no_lab" |
    Case_aux c "eval_label_of" |
    Case_aux c "eval_get_pc" |
    Case_aux c "eval_mk_nav" |
    Case_aux c "eval_to_sum"
  ].

Tactic Notation "evalLT_cases" tactic(first) ident(c) :=
  first; [
  (* perl -ne 'print "    Case_aux c \"$1\" |\n" if /\| (eval_\w+)\s/' < LThrowBigStep.v *)
    Case_aux c "eval_var" |
    Case_aux c "eval_const" |
    Case_aux c "eval_let" |
    Case_aux c "eval_let_fail" |
    Case_aux c "eval_abs" |
    Case_aux c "eval_app" |
    Case_aux c "eval_app_etype" |
    Case_aux c "eval_inx" |
    Case_aux c "eval_match" |
    Case_aux c "eval_match_etype" |
    Case_aux c "eval_tag" |
    Case_aux c "eval_bop" |
    Case_aux c "eval_bracket" |
    Case_aux c "eval_bracket_etype" |
    Case_aux c "eval_label_of" |
    Case_aux c "eval_get_pc" |
    Case_aux c "eval_throw" |
    Case_aux c "eval_catch_no_excp" |
    Case_aux c "eval_catch_excp"
  ].


(* TODO: extract as many of the following definitions as possible and
         use them for QuickChecking *)

Fixpoint InLN (t : Tm) : Prop :=
  match t with
  | (TThrow _) => False
  | (TCatch _ _ _) => False
  | (TLet _ t1 t2) => InLN t1 /\ InLN t2
  | (TAbs _ t) => InLN t
  | (TMatch _ _ t1 t2) => InLN t1 /\ InLN t2
  | (TBracket _ t) => InLN t
  | _ => True
  end.

(* Equality relation on atoms / values/ terms
   - up to alpha equivalence (m stores a variable renaming)
   - comparing closures makes everything much more complicated
     since we need to treat the captured environment as delayed
     substitutions that can affect the equality of open terms
     (r1 and r2 in eq_tmLB_mr and eq_varLB_mr)
   - defining this as an inductive relation instead of a function
     (there seems to be no easy way around that --
     it seems impossible to convince Coq that this is terminating) *)
Inductive eq_varLB_mr : list (Var * Var) ->
                        LBracketSyntax.Env ->
                        LBracketSyntax.Env ->
                        Var -> Var -> Prop :=
  | EqX1 : forall m r1 r2 x1 x2, (* bound variables *)
    In (x1, x2) m ->
    eq_varLB_mr m r1 r2 x1 x2
  | EqX2 : forall m r1 r2 x1 x2 a1 a2, (* free variables *)
    ~In (x1, x2) m ->
    In (x1,a1) r1 ->
    In (x2,a2) r2 ->
    (* r1 and r2 not passed to eq_atomLB_m; the atoms in r1/r2 are
    self-contained -- i.e. each closure carries its own environment *)
    eq_atomLB_m [] a1 a2 -> 
    eq_varLB_mr m r1 r2 x1 x2
  | EqXRefl : forall m r1 r2 x,
    eq_varLB_mr m r1 r2 x x
with eq_tmLB_mr : list (Var * Var) ->
                  LBracketSyntax.Env ->
                  LBracketSyntax.Env ->
                  Tm -> Tm -> Prop :=
  | EqTVar : forall m r1 r2 x1 x2,
    eq_varLB_mr m r1 r2 x1 x2 ->
    eq_tmLB_mr m r1 r2 (TVar x1) (TVar x2)
  | EqTConst : forall m r1 r2 c,
    eq_tmLB_mr m r1 r2 (TConst c) (TConst c)
  | EqTLet : forall m r1 r2 x1 t1' t1'' x2 t2' t2'',
    eq_tmLB_mr m r1 r2 t1' t2' ->
    eq_tmLB_mr ((x1,x2)::m) r1 r2 t1'' t2'' ->
    eq_tmLB_mr m r1 r2 (TLet x1 t1' t1'') (TLet x2 t2' t2'')
  | EqTAbs : forall m r1 r2 x1 t1' x2 t2',
    eq_tmLB_mr ((x1,x2)::m) r1 r2 t1' t2' ->
    eq_tmLB_mr m r1 r2 (TAbs x1 t1') (TAbs x2 t2')
  | EqTApp : forall m r1 r2 x1 x1' x2 x2',
    eq_varLB_mr m r1 r2 x1 x2 ->
    eq_varLB_mr m r1 r2 x1' x2' ->
    eq_tmLB_mr m r1 r2 (TApp x1 x1') (TApp x2 x2')
  | EqTInx : forall m r1 r2 d x1 x2,
    eq_varLB_mr m r1 r2 x1 x2 ->
    eq_tmLB_mr m r1 r2 (TInx d x1) (TInx d x2)
  | EqTMatch : forall m r1 r2 x1 x1' t1' t1'' x2 x2' t2' t2'',
    eq_varLB_mr m r1 r2 x1 x2 ->
    eq_tmLB_mr ((x1',x2')::m) r1 r2 t1' t2' ->
    eq_tmLB_mr ((x1',x2')::m) r1 r2 t1'' t2'' ->
    eq_tmLB_mr m r1 r2 (TMatch x1 x1' t1' t1'') (TMatch x2 x2' t2' t2'')
  | EqTTag : forall m r1 r2 x1 x2,
    eq_varLB_mr m r1 r2 x1 x2 ->
    eq_tmLB_mr m r1 r2 (TTag x1) (TTag x2)
  | EqTBOp : forall m r1 r2 bo x1 x1' x2 x2',
    eq_varLB_mr m r1 r2 x1 x2 ->
    eq_varLB_mr m r1 r2 x1' x2' ->
    eq_tmLB_mr m r1 r2 (TBOp bo x1 x1') (TBOp bo x2 x2')
  | EqTBracket : forall m r1 r2 x1 t1' x2 t2',
    eq_varLB_mr m r1 r2 x1 x2 ->
    eq_tmLB_mr m r1 r2 t1' t2' ->
    eq_tmLB_mr m r1 r2 (TBracket x1 t1') (TBracket x2 t2')
  | EqTLabelOf : forall m r1 r2 x1 x2,
    eq_varLB_mr m r1 r2 x1 x2 ->
    eq_tmLB_mr m r1 r2 (TLabelOf x1) (TLabelOf x2)
  | EqTGetPc : forall m r1 r2,
    eq_tmLB_mr m r1 r2 TGetPc TGetPc
  | EqTMkNav : forall m r1 r2 x1 x2,
    eq_varLB_mr m r1 r2 x1 x2 ->
    eq_tmLB_mr m r1 r2 (TMkNav x1) (TMkNav x2)
  | EqTToSum : forall m r1 r2 x1 x2,
    eq_varLB_mr m r1 r2 x1 x2 ->
    eq_tmLB_mr m r1 r2 (TToSum x1) (TToSum x2)
  | EqTThrow : forall m r1 r2 x1  x2,
    eq_varLB_mr m r1 r2 x1 x2 ->
    eq_tmLB_mr m r1 r2 (TThrow x1) (TThrow x2)
  | EqTCatch : forall m r1 r2 t1' x1 t1'' t2' x2 t2'',
    eq_tmLB_mr m r1 r2 t1' t2' ->
    eq_tmLB_mr ((x1,x2)::m) r1 r2 t1'' t2'' ->
    eq_tmLB_mr m r1 r2 (TCatch t1' x1 t1'') (TCatch t2' x2 t2'')
with eq_atomLB_m : list (Var * Var) ->
                   LBracketSyntax.Atom ->
                   LBracketSyntax.Atom -> Prop :=
  | EqAtom : forall m l v1 v2,
    eq_valLB_m m v1 v2 ->
    eq_atomLB_m m (v1,l) (v2,l)
with eq_valLB_m  : list (Var * Var) ->
                    LBracketSyntax.Val ->
                    LBracketSyntax.Val -> Prop :=
  | EqVConst : forall m c,
    eq_valLB_m m (LBracketSyntax.VConst c) (LBracketSyntax.VConst c)
  | EqVInx : forall m d a1 a2,
    eq_atomLB_m m a1 a2 ->
    eq_valLB_m m (LBracketSyntax.VInx d a1) (LBracketSyntax.VInx d a2)
  | EqVClos : forall m r1 x1 t1 r2 x2 t2,
    eq_tmLB_mr ((x1, x2)::m) r1 r2 t1 t2 ->
    eq_valLB_m m (LBracketSyntax.VClos r1 x1 t1)
                 (LBracketSyntax.VClos r2 x2 t2).

Hint Constructors eq_valLB_m eq_atomLB_m eq_tmLB_mr.

Scheme eq_varLB_mr_ind' := Minimality for eq_varLB_mr Sort Prop
with eq_tmLB_mr_ind' := Minimality for eq_tmLB_mr Sort Prop
with eq_atomLB_m_ind' := Minimality for eq_atomLB_m Sort Prop
with eq_valLB_m_ind' := Minimality for eq_valLB_m Sort Prop.

Combined Scheme eqLB_mutind from 
  eq_varLB_mr_ind', eq_tmLB_mr_ind', eq_atomLB_m_ind', eq_valLB_m_ind'.

(* TODO: this looks too strong to me now -- no weakening rule?
   same variable name? *)
Inductive eqEnvLB : LBracketSyntax.Env ->
                    LBracketSyntax.Env -> Prop :=
  | EqNil :
    eqEnvLB [] []
  | EqCons : forall x a1 r1 a2 r2,
    eq_atomLB_m [] a1 a2 ->
    eqEnvLB r1 r2 ->
    eqEnvLB ((x,a1) :: r1) ((x,a2) :: r2).

(* Extending the translation from LNaV to LThrow to values, atoms,
   and environments *)
(* TODO: extract this and make sure it works *)
Definition Pval' (v : LNaVSyntax.Val) := LBracketSyntax.Val.
Definition Pbox' (b : LNaVSyntax.Box) := LBracketSyntax.Val.
Definition Patom' (a : LNaVSyntax.Atom) := LBracketSyntax.Atom.
Definition Penv' (r : LNaVSyntax.Env) := LBracketSyntax.Env.
Definition H_v_const (c : Const) : LBracketSyntax.Val :=
  LBracketSyntax.vInl (LBracketSyntax.VConst c, bot).
Definition H_v_inx (d : Dir) (_aLN : LNaVSyntax.Atom)
    (HaLN : LBracketSyntax.Atom) : LBracketSyntax.Val :=
  LBracketSyntax.vInl (LBracketSyntax.VInx d HaLN, bot).
Definition H_v_clos (_r : LNaVSyntax.Env) (x : Var) (t : Tm)
    (Hr : LBracketSyntax.Env) : LBracketSyntax.Val :=
  LBracketSyntax.vInl (LBracketSyntax.VClos Hr x (encode t) , bot).
Definition H_b_val (v : LNaVSyntax.Val)
    (Hv : LBracketSyntax.Val) : LBracketSyntax.Val := Hv.
Definition H_b_nav (e : Excp) : LBracketSyntax.Val :=
  LBracketSyntax.vInr (LBracketSyntax.vExcp e, bot).
Definition H_atom (b : LNaVSyntax.Box) (l : Lab)
    (Hb : LBracketSyntax.Val) : LBracketSyntax.Atom := (Hb, l).
Definition H_nil : LBracketSyntax.Env := [].
Definition H_cons (x : Var) (_a : LNaVSyntax.Atom) (_r : LNaVSyntax.Env)
    (Ha : LBracketSyntax.Atom) (Hr : LBracketSyntax.Env)
    : LBracketSyntax.Env := (x,Ha) :: Hr.

Definition atomEN2T : LNaVSyntax.Atom -> LBracketSyntax.Atom :=
  LNaVSyntax.atom_rect Pval' Pbox' Patom' Penv'
    H_v_const H_v_inx H_v_clos H_b_val H_b_nav H_atom H_nil H_cons.

Definition valEN2T : LNaVSyntax.Val -> LBracketSyntax.Val :=
  LNaVSyntax.val_rect Pval' Pbox' Patom' Penv'
    H_v_const H_v_inx H_v_clos H_b_val H_b_nav H_atom H_nil H_cons.

Definition envEN2T : LNaVSyntax.Env -> LBracketSyntax.Env :=
  LNaVSyntax.env_rect Pval' Pbox' Patom' Penv'
    H_v_const H_v_inx H_v_clos H_b_val H_b_nav H_atom H_nil H_cons.

Lemma atomEN2T_eq : forall a,
  atomEN2T a =
    match a with
    | (LNaVSyntax.V v,l) =>
        (valEN2T v, l)
    | (LNaVSyntax.D e,l) =>
        (LBracketSyntax.vInr (LBracketSyntax.vExcp e, bot), l)
    end.
Proof. intros. destruct a as [[v | e] l]; reflexivity. Qed.

Lemma envEN2T_eq : forall rho,
  envEN2T rho = match rho with
                | [] => []
                | (x,a) :: rho => (x, atomEN2T a) :: envEN2T rho
                end.
Proof. intros. destruct rho as [| [x a] rho']; reflexivity. Qed.

Lemma envEN2T_eq' : forall rho,
  envEN2T rho = map (fun xa => match xa with
                               (x,a) => (x, atomEN2T a)
                               end) rho.
Proof. induction rho as [| [x a] rho']; simpl. reflexivity.
  rewrite IHrho'. reflexivity. Qed.

Lemma valEN2T_eq : forall v,
  valEN2T v =
    LBracketSyntax.vInl
      (match v with
      | LNaVSyntax.VConst c =>
          LBracketSyntax.VConst c
      | LNaVSyntax.VInx d aLN =>
          LBracketSyntax.VInx d (atomEN2T aLN)
      | LNaVSyntax.VClos rho x t =>
          LBracketSyntax.VClos (envEN2T rho) x (encode t)
      end, bot).
Proof.
  apply LNaVSyntax.val_ind with
    (Penv := fun rho =>
      envEN2T rho = match rho with
                    | [] => []
                    | (x,a) :: rho => (x, atomEN2T a) :: envEN2T rho
                    end)
     (Patom := fun a => True) (Pbox := fun b => True).
  intros. reflexivity.
  intros. destruct a as [[ | ] l]; reflexivity.
  Case "the only interesting case".
    intros. rewrite H. clear H. simpl.
    destruct r as [| [x' [b l]] r']. reflexivity.
    rewrite atomEN2T_eq. rewrite envEN2T_eq. simpl.
    unfold H_atom, H_b_val.
    unfold H_v_clos, H_cons.
  Lemma helper : forall r',
  (fix env_rect' (r : LNaVSyntax.Env) : Penv' r :=
                 match r as r0 return (Penv' r0) with
                 | nil => H_nil
                 | p :: r'0 =>
                     let (x0, a) as p0 return (Penv' (p0 :: r'0)) := p in
                     (x0,
                     (fix atom_rect' (a0 : LNaVSyntax.Atom) : 
                      Patom' a0 :=
                        let (b0, l0) as a1 return (Patom' a1) := a0 in
                        match b0 as b1 return (Patom' (b1, l0)) with
                        | LNaVSyntax.V v => (valEN2T v, l0)
                        | LNaVSyntax.D e => (H_b_nav e, l0)
                        end) a) :: env_rect' r'0
                 end) r'
  = match r' with
              | nil => []
              | (x0, a) :: rho => (x0, atomEN2T a) :: envEN2T rho
              end.
  Proof.
    induction r' as [| [x'' [b'' l'']] r'']; intros. reflexivity.
    rewrite IHr''. clear IHr''. rewrite atomEN2T_eq. rewrite envEN2T_eq.
    reflexivity.
  Qed.
  rewrite helper. reflexivity.
  trivial. trivial. trivial. trivial. trivial.
Qed.

Lemma maps_envEN2T : forall rLN y aLN rLT,
  maps rLN y aLN ->
  eqEnvLB (envEN2T rLN) rLT ->
  exists aLT, maps rLT y aLT /\ 
    eq_atomLB_m [] (atomEN2T aLN) aLT.
Proof.
  intros. remember (envEN2T rLN).
  generalize dependent rLN. generalize dependent y. generalize dependent aLN.
  induction H0; intros; destruct rLN as [| [x' a] rLN'].
    inversion H. inversion Heqe. inversion Heqe. 
    rewrite envEN2T_eq in Heqe. invsc Heqe.
    invsc H1.
    remember (beq_nat y x') as b. destruct b.
    Case "head".
      apply beq_nat_eq in Heqb. invsc H3.
      eexists. split; [| eassumption].
        unfold maps. simpl. rewrite <- beq_nat_refl. reflexivity.
    Case "tail".
      apply IHeqEnvLB in H3; [| reflexivity].
      destruct H3 as [aLT [H31 H32]].
      eexists. split; [| eassumption].
        unfold maps. simpl. rewrite <- Heqb. apply H31.
Qed.      

Lemma eq_var2_dec : forall p1 p2 : Var * Var, {p1 = p2} + {p1 <> p2}.
Proof. intros. decide equality; apply eq_var_dec. Qed.

Lemma eq_tmLB_mr_refl :
  forall t m r1 r2, eq_tmLB_mr m r1 r2 t t.
Proof. (Tm_cases (induction t) Case); intros; eauto using EqXRefl.
  (* Went for the simple fix and added an additonal reflexivity rule
     to eq_varLB_mr -- not sure if that's the correct long term
     solution; we might need to additional well-formedness conditions
     on r1 r2 and t in the main theorem *)
Qed.

Lemma eq_atomLB_m_refl :
  forall a m, eq_atomLB_m m a a.
Admitted. (* XXX *)

Lemma maps_cons : forall X (x : Var) (a : X) r,
  maps ((x, a) :: r) x a.
Proof. intros. unfold maps. simpl. rewrite <- beq_nat_refl. reflexivity. Qed.

Lemma maps_cons2 : forall X (x : Var) (a : X) (y : Var) (a' : X) r,
  beq_nat x y = false ->
  maps ((y,a') :: (x, a) :: r) x a.
Proof. intros. unfold maps. simpl. rewrite H.
  rewrite <- beq_nat_refl. reflexivity. Qed.

Lemma maps_cons3 : forall X (x : Var) (a : X) (y : Var) (ay : X) 
    (z : Var) (az : X)r,
  beq_nat x y = false ->
  beq_nat x z = false ->
  maps ((z,az) :: (y,ay) :: (x, a) :: r) x a.
Proof. intros. unfold maps. simpl. rewrite H. rewrite H0.
  rewrite <- beq_nat_refl. reflexivity. Qed.

Lemma maps_tail : forall X x y (a ay : X) r,
  beq_nat x y = false ->
  maps r x a ->
  maps ((y, ay) :: r) x a.
Proof. intros. unfold maps. simpl. rewrite H. apply H0. Qed.

Lemma five_rel_eq_four : forall (X1 X2 X3 X4 X5: Type)
  (R : X1 -> X2 -> X3 -> X4 -> X5 -> Prop) x1 x2 x3 x4 x4' x5,
  R x1 x2 x3 x4 x5 -> x4 = x4' -> R x1 x2 x3 x4' x5.
Proof. intros. subst. assumption. Qed.

Lemma if_eq_var_dec_refl : forall x X (a b : X),
  (if eq_var_dec x x then a else b) = a.
Proof. intros. destruct eq_var_dec; tauto. Qed.

(* Note: a lot of the trouble below comes from the usage of the
   environmental big-step semantics. I think that the proof would be a
   lot easier with a substitution-based big-step semantics. Atoms
   would need to be terms, but that won't interact well with our ANF
   variant (currently only Inl x and Inr x are terms). Q: Do we really
   need to make atoms _be_ terms? Isn't it enough if atoms can be
   embedded into terms?

  [[c]]_v = c
  [[Inx a]]_v = let z = [[a]]_a in Inx z
  [[\x.t]]_v = \x.t
  [[v@L]]_a = L<v>
  [(D e)@L]]_a = L<mkNaV e> -- for LNaV + LThrowD atoms are complicated

Something along these lines:

Fixpoint val_to_tm (v:Val) : Tm :=
  match v with
  | VConst c => TConst c
  | VInx d a => tInx d (atom_to_tm a)
  | VFun x t => TFun x t
  end
with atom_to_tm (a:Atom) : Tm :=
  match a with
  | (V v, l) => tBracket (tCLab l) (val_to_tm v)
  | (D e, l) => tBracket (tCLab l) (tMkNav (tCExcp e))
  end.

The property we would want is that "atom_to_tm a" evaluates to a.

The inductive invariants needed when using this semantics would
probably be simpler, but we risk opening another pandora box:
substitution. *)

(* XXX very far from the truth *)
Lemma helper' : forall x x2 r1 r2 t1 t2 pc pc' aLT aLT',
  eq_tmLB_mr [(x, x2)] r1 r2 t1 t2 ->
  LThrowBigStep.eval ((x, aLT) :: r1) t1 pc (LThrowBigStep.Suc aLT') pc' ->
  exists aLT'',
  LThrowBigStep.eval ((x2, aLT) :: r2) t2 pc (LThrowBigStep.Suc aLT'') pc'
  /\ eq_atomLB_m [] aLT' aLT''.
Proof.
  intros. generalize dependent r2. generalize dependent x2.
  generalize dependent t2. remember ((x, aLT) :: r1) as r.
  generalize dependent x. generalize dependent aLT.
  generalize dependent r1. remember (LThrowBigStep.Suc aLT').
  generalize dependent aLT'.
  (evalLT_cases (induction H0) Case); intros; subst.
  Case "eval_var". invsc H0. invsc Heqr0.
    eexists. split. econstructor. invsc H5. (* 3 possible subcases *)
    SCase "EqX1".
      simpl in H0. invsc H0; [| intuition]. invsc H1.
      unfold maps. simpl. rewrite <- beq_nat_refl. reflexivity.
    SCase "EqX2".
      simpl in H0. unfold maps in H. simpl in H.
      skip. (* ??? *)
    SCase "EqXRefl". skip.
      (* XXX this doesn't hold -- so we need extra premises or
             we need to drop this rule from eq_varLB_mr *)
Admitted. (* XXX very far from the truth *)

Lemma eq_atomLB_m_trans : forall m a1 a2 a3,
  eq_atomLB_m m a1 a2 ->
  eq_atomLB_m m a2 a3 ->
  eq_atomLB_m m a1 a3.
Admitted.

Lemma eqEnvLB_refl : forall r,
  eqEnvLB r r.
Proof.
  induction r as [| [x a] r']; constructor.
  apply eq_atomLB_m_refl. assumption. 
Qed.

Inductive InLNval : LNaVSyntax.Val -> Prop :=
  | InLN_clos : forall r x t, 
    InLN t ->
    InLNenv r ->
    InLNval (LNaVSyntax.VClos r x t)
  | InLN_const : forall c, 
    InLNval (LNaVSyntax.VConst c)
  | InLN_inx : forall d a, 
    InLNatom a ->
    InLNval (LNaVSyntax.VInx d a)
with InLNatom : LNaVSyntax.Atom -> Prop :=
  | InLN_val : forall v l, 
    InLNval v ->
    InLNatom (LNaVSyntax.V v, l)
  | InLN_nav : forall e l, 
    InLNatom (LNaVSyntax.D e, l)
with InLNenv : LNaVSyntax.Env -> Prop :=
  | InLN_nil : InLNenv []
  | InLN_cons : forall x a r,
    InLNenv r ->
    InLNatom a ->
    InLNenv ((x,a) :: r).

Lemma eval_preserves_InLN : forall r t pc a pc',
  InLN t ->
  LNaVBigStep.eval r t pc a pc' ->
  InLNatom a.
Admitted.

Lemma InLNenv_maps : forall r x a,
  InLNenv r ->
  maps r x a ->
  InLNatom a.
Admitted.

(* Because we are using the big-step semantics this only talks about
   successful runs (non-termination ignored) *)
Theorem prop_EN2T_correct : forall tLN rLN pc aLN pc' rLT,
  InLN tLN ->
  InLNenv rLN ->
  LNaVBigStep.eval rLN tLN pc aLN pc' ->
  eqEnvLB (envEN2T rLN) rLT ->
  exists aLT,
  LThrowBigStep.eval rLT (encode tLN) pc (LThrowBigStep.Suc aLT) pc'
  /\ eq_atomLB_m [] (atomEN2T aLN) aLT.
Proof.
  intros tLN rLN pc aLN pc' rLT HInLN HInLNenv Heval Henv.
  generalize dependent rLT.
  (evalLN_cases (induction Heval) Case); intros.
  Case "eval_var".
    simpl. eapply maps_envEN2T in Henv.
    destruct Henv as [aLT [G1 G2]]. eauto. trivial.
  Case "eval_const".
    exists (LBracketSyntax.vInl (LBracketSyntax.VConst c, bot), bot).
    split.
    SCase "Left conjunct".
      unfold tInl. eapply LThrowBigStep.eval_let. eauto.
      apply LThrowBigStep.eval_inx. unfold maps. simpl. reflexivity.
    SCase "Right conjunct". 
     repeat constructor.
  Case "eval_let".
    simpl in *. destruct HInLN as [G1 G2].
    specialize (IHHeval1 G1 HInLNenv _ Henv).
      destruct IHHeval1 as [aLT [IH11 IH12]].
    assert (eqEnvLB (envEN2T ((x, a) :: r)) ((x, aLT) :: rLT)) as G.
      rewrite envEN2T_eq. apply EqCons; assumption.
    assert(InLNenv ((x, a) :: r)) as GInLNenv.
      constructor. assumption. eauto using eval_preserves_InLN.
    specialize (IHHeval2 G2 GInLNenv _ G).
      destruct IHHeval2 as [aLT' [IH21 IH22]].
    exists aLT'. split.
    SCase "Left conjunct". simpl.
      eapply LThrowBigStep.eval_let; eassumption.
    SCase "Right conjunct". assumption.
  Case "eval_abs".
    exists (LBracketSyntax.vInl
      (LBracketSyntax.VClos rLT x (encode t), bot), bot).
    split.
      simpl. repeat econstructor. unfold LBracketSyntax.vInl.
        rewrite atomEN2T_eq. rewrite valEN2T_eq.
        repeat constructor.
      apply eq_tmLB_mr_refl.
  Case "eval_app".

    assert(InLNatom a) as HInLNa by eauto using InLNenv_maps.

    assert(InLN t) as GInLNt.
      apply InLNenv_maps in H. invsc H. invsc H2. assumption. assumption.

    assert(InLNenv ((x, a) :: r')) as GInLNenv.
      apply InLNenv_maps in H. invsc H. invsc H2.
      constructor. assumption. assumption. assumption.

    eapply maps_envEN2T in H; [| eassumption].
    destruct H as [aLT0 [H11 H12]]. rewrite atomEN2T_eq in H12.
    rewrite valEN2T_eq in H12.
    invsc H12. invsc H4. invsc H5. invsc H4. (* these generate t2 *)

    eapply maps_envEN2T in H0; [| eassumption].
    destruct H0 as [aLT [H31 H32]].

    assert(eqEnvLB (envEN2T ((x, a) :: r')) ((x, aLT) :: envEN2T r')) as Geq.
      rewrite envEN2T_eq. constructor. assumption.
      apply eqEnvLB_refl.

    specialize (IHHeval GInLNt GInLNenv _ Geq).
    edestruct IHHeval as [aLT' [IH11 IH12]].
    eapply helper' in IH11; [| eassumption].
    destruct IH11 as [aLT'' [IH111 IH112]].

    eexists. split.
    SSCase "Left conjunct (evaluation)".
      simpl. unfold tSBind, tMatch.

      eapply LThrowBigStep.eval_let with
        (a := (LBracketSyntax.VInx DLeft
          (LBracketSyntax.VClos r2 x2 t2, bot), l))
        (pc' := pc).
      constructor. apply H11.
      econstructor.
      apply maps_cons. unfold tHasTag.
      econstructor. econstructor. repeat econstructor.
      apply maps_cons. apply maps_cons. econstructor.
      constructor. eapply five_rel_eq_four.
      apply LThrowBigStep.eval_bop.
      apply maps_cons2. simpl. rewrite if_eq_var_dec_refl. simpl.
        skip. (* XXX beq_nat + fresh *)
      apply maps_cons. simpl. reflexivity.
      eapply LThrowBigStep.eval_match.
      apply maps_cons. unfold d_choose. econstructor.
      apply maps_cons3. simpl. rewrite if_eq_var_dec_refl.
        skip. (* XXX beq_nat + fresh *)
        simpl. rewrite if_eq_var_dec_refl.
        skip. (* XXX beq_nat + fresh *)

        apply maps_tail. skip. (* XXX beq_nat + fresh *)
        apply maps_tail. skip. (* XXX beq_nat + fresh *)
        apply maps_tail. skip. (* XXX beq_nat + fresh *)
        apply maps_tail. skip. (* XXX beq_nat + fresh *)
        eassumption.

      repeat rewrite join_bot. apply IH111.
    SSCase "Right conjunct".
      eapply eq_atomLB_m_trans; eassumption.
  Case "eval_app_no_abs". admit.
  Case "eval_inx". admit.
  Case "eval_match". admit.
  Case "eval_match_no_sum". admit.
  Case "eval_tag". admit.
  Case "eval_bop". admit.
  Case "eval_bracket". admit.
  Case "eval_bracket_no_lab". admit.
  Case "eval_label_of". admit.
  Case "eval_get_pc".
    eexists. split.
    SCase "Left conjunct". simpl. econstructor. eauto.
      econstructor. unfold maps. simpl. reflexivity.
    SCase "Right conjunct". repeat constructor.
  Case "eval_mk_nav". admit.
  Case "eval_to_sum". admit.
Admitted.
